r"""Utils
=========
"""
import json
import os
from typing import Any, Literal

from tree_sitter.binding import Query
from tree_sitter_languages import get_language

from . import FILETYPE

SCHEMAS = {}
QUERIES = {}


def get_query(name: str, filetype: str = "bash") -> Query:
    r"""Get query.

    :param name:
    :type name: str
    :param filetype:
    :type filetype: str
    :rtype: Query
    """
    if name not in QUERIES:
        with open(
            os.path.join(
                os.path.join(
                    os.path.join(os.path.dirname(__file__), "assets"),
                    "queries",
                ),
                f"{name}{os.path.extsep}scm",
            )
        ) as f:
            text = f.read()
        language = get_language(filetype)
        QUERIES[name] = language.query(text)
    return QUERIES[name]


def get_schema(filetype: FILETYPE) -> dict[str, Any]:
    r"""Get schema.

    :param filetype:
    :type filetype: FILETYPE
    :rtype: dict[str, Any]
    """
    if filetype not in SCHEMAS:
        file = os.path.join(
            os.path.join(
                os.path.join(os.path.dirname(__file__), "assets"),
                "json",
            ),
            f"{filetype}.json",
        )
        with open(file, "r") as f:
            SCHEMAS[filetype] = json.load(f)
    return SCHEMAS[filetype]


def get_filetype(uri: str) -> FILETYPE | Literal[""]:
    r"""Get filetype.

    :param uri:
    :type uri: str
    :rtype: FILETYPE | Literal[""]
    """
    basename = os.path.basename(uri)
    ext = uri.split(os.path.extsep)[-1]
    if basename == "build.sh":
        return "build.sh"
    if basename.endswith(".subpackage.sh"):
        return "subpackage.sh"
    if ext == "install":
        return "install"
    if basename == "PKGBUILD":
        return "PKGBUILD"
    if basename == "makepkg.conf":
        return "makepkg.conf"
    if ext in {"ebuild", "eclass"}:
        return "ebuild"
    if basename == "make.conf":
        return "make.conf"
    if basename == "color.map":
        return "color.map"
    return ""



r"""Packages
============
"""
from .. import FILETYPE

PACKAGE_VARIABLES = {
    "PKGBUILD": {
        "depends",
        "makedepends",
        "optdepends",
        "conflicts",
        "provides",
        "replaces",
    }
}


def search_package_document(name: str, filetype: FILETYPE) -> str:
    r"""Search package document.

    :param name:
    :type name: str
    :param filetype:
    :type filetype: FILETYPE
    :rtype: str
    """
    if filetype == "PKGBUILD":
        from .pkgbuild import get_package_document
    else:
        raise NotImplementedError
    return get_package_document(name)


def search_package_names(name: str, filetype: FILETYPE) -> dict[str, str]:
    r"""Search package names.

    :param name:
    :type name: str
    :param filetype:
    :type filetype: FILETYPE
    :rtype: dict[str, str]
    """
    if filetype == "PKGBUILD":
        from .pkgbuild import get_package_names
    else:
        raise NotImplementedError
    return get_package_names(name)



r"""Licenses
============
"""
from license_expression import get_license_index

LICENSES = [
    i["spdx_license_key"]
    for i in get_license_index()
    if i.get("spdx_license_key")
    and not i["spdx_license_key"].startswith("LicenseRef-scancode-")
]
ATOM = f"({'|'.join(LICENSES)})"



r"""Test utils."""
from termux_language_server.utils import get_schema


class Test:
    r"""Test."""

    @staticmethod
    def test_get_schema() -> None:
        r"""Test get schema.

        :rtype: None
        """
        assert len(
            get_schema("build.sh")
            .get("properties", {})
            .get("TERMUX_PKG_VERSION", {})
            .get("description", "")
            .splitlines()
        )



r"""PKGBUILD
============
"""
import os
from typing import Any

from markdown_it.token import Token
from tree_sitter_lsp.misc import get_md_tokens, get_soup

from .._metainfo import SOURCE, project
from .licenses import LICENSES


def get_content(tokens: list[Token]) -> str:
    r"""Get content.

    :param tokens:
    :type tokens: list[Token]
    :rtype: str
    """
    return "\n".join(
        [token.content.replace("\n", " ") for token in tokens if token.content]
    )


def init_schema() -> dict[str, Any]:
    r"""Init schema.

    :rtype: dict[str, dict[str, Any]]
    """
    schemas = {}
    for filetype in {"PKGBUILD", "install"}:
        schemas[filetype] = {
            "$id": f"{SOURCE}/blob/main/src/termux_language_server/assets/json/{filetype}.json",
            "$schema": "http://json-schema.org/draft-07/schema#",
            "$comment": (
                "Don't edit this file directly! It is generated by "
                f"`{project} --generate-schema={filetype}`."
            ),
            "type": "object",
            "patternProperties": {},
        }
    # PKGBUILD has variables
    schemas["PKGBUILD"]["properties"] = {}

    tokens = get_md_tokens("PKGBUILD")
    # **pkgname (array)**
    #
    # > Either the name of the package or an array of names for split
    # > packages. Valid characters for members of this array are
    # > alphanumerics, and any of the following characters: "@ . \_ + -".
    # > Additionally, names are not allowed to start with hyphens or dots.
    indices = [
        index
        for index, token in enumerate(tokens)
        if token.content.startswith("**")
        and token.content.endswith("**")
        and token.level == 1
    ]
    blockquote_close_indices = [
        index
        for index, token in enumerate(tokens)
        if token.type == "blockquote_close"
    ]
    close_indices = [
        min(
            [
                blockquote_close_index
                for blockquote_close_index in blockquote_close_indices
                if blockquote_close_index > index
            ]
        )
        for index in indices
    ]
    for index, close_index in zip(indices, close_indices):
        children = tokens[index].children
        if children is None:
            continue
        words = [child.rstrip(",") for child in children[2].content.split()]
        # md5sums, ...
        if len(words) > 2:
            _words = [child for child in children[4].content.split()]
            kind = (
                _words[1].lstrip("(").rstrip(")")
                if len(words) > 1
                else "string"
            )
            names = words + [_words[0]]
        else:
            names = [words[0].rstrip("()")]
            kind = (
                words[1].lstrip("(").rstrip(")")
                if len(words) > 1
                else "string"
            )
        for name in names:
            description = get_content(tokens[index + 1 : close_index])
            if name.startswith("pre_") or name.startswith("post_"):
                filetype = "install"
                kind = "Function"
            else:
                filetype = "PKGBUILD"
            if kind == "Function":
                name = rf"^{name}($|_.*)"
                properties_name = "patternProperties"
            else:
                properties_name = "properties"
            schemas[filetype][properties_name][name] = {
                "description": description
            }
            # makepkg supports building multiple packages from a single PKGBUILD.
            # This is achieved by assigning an array of package names to the
            # pkgname directive.
            if name == "pkgname":
                schemas[filetype][properties_name][name]["oneOf"] = [
                    {
                        "type": "array",
                        "items": {"type": "string"},
                        "uniqueItems": True,
                    },
                    {"type": "string"},
                ]
            elif kind == "string":
                schemas[filetype][properties_name][name]["type"] = "string"
            elif kind in {"array", "arrays"}:
                schemas[filetype][properties_name][name] |= {
                    "type": "array",
                    "items": {"type": "string"},
                    "uniqueItems": True,
                }
            elif kind == "Function":
                # Each split package uses a corresponding packaging function with
                # name package_foo(), where foo is the name of the split package.
                schemas[filetype][properties_name][name]["const"] = 0
    # https://archlinux32.org/architecture/
    # https://archlinux.org/packages/
    # https://archlinuxarm.org/forum/viewforum.php?f=56
    schemas["PKGBUILD"]["properties"]["arch"]["items"]["enum"] = [
        "any",
        "pentium4",
        "i486",
        "i686",
        "x86_64",
        "x86_64_v3",
        "arm",
        "armv6h",
        "armv7h",
        "armv8",
        "aarch64",
    ]
    schemas["PKGBUILD"]["properties"]["url"]["format"] = "uri"
    del schemas["PKGBUILD"]["properties"]["pkgver"]["type"]
    schemas["PKGBUILD"]["properties"]["pkgver"] |= {
        "oneOf": [{"type": "string"}, {"const": 0}]
    }

    soup = get_soup("https://www.msys2.org/dev/pkgbuild/")
    for tr in soup.find_all("tr")[1:]:
        tds = tr.find_all("td")
        name = tds[0].code.text
        kind = tds[1].text
        kind = {"mapping": "array"}.get(kind, kind)
        description = tds[2].text
        schemas["PKGBUILD"]["properties"][name] = {
            "type": kind,
            "description": description,
        }
        if kind == "array":
            schemas["PKGBUILD"]["properties"][name] |= {
                "items": {"type": "string"},
                "uniqueItems": True,
            }
        elif kind == "object":
            schemas["PKGBUILD"]["properties"][name] |= {"properties": {}}
        elif kind == "string" and name.endswith("_url"):
            schemas["PKGBUILD"]["properties"][name]["format"] = "uri"
    # https://packages.msys2.org/repos
    schemas["PKGBUILD"]["properties"]["mingw_arch"]["items"]["enum"] = [
        "mingw32",
        "mingw64",
        "ucrt64",
        "clang64",
        "clang32",
        "clangarm64",
    ]
    names = []
    for li in soup.find_all("li"):
        code = li.find("code")
        if code is None:
            continue
        name = code.text
        if not li.text.startswith(name):
            continue
        names += [name]
        # text = li.text
        # _, _, text = text.partition(name)
        # description = text.replace("\n", " ").lstrip("- ")
    schemas["PKGBUILD"]["properties"]["msys2_references"]["items"][
        "pattern"
    ] = f"({'|'.join(names)})(|: .*)"
    schemas["PKGBUILD"]["properties"]["license"]["items"] = {
        "oneOf": [
            {
                "type": "string",
                "enum": LICENSES
                + os.listdir("/usr/share/licenses/common")
                + ["custom"],
            },
            {"type": "string", "pattern": "custom:.+"},
        ]
    }
    return schemas



r"""Misc
========
"""
from typing import Any

from .. import FILETYPE


def get_schema(filetype: FILETYPE) -> dict[str, Any]:
    r"""Get schema.

    :param filetype:
    :type filetype: FILETYPE
    :rtype: dict[str, Any]
    """
    if filetype in {"build.sh", "subpackage.sh"}:
        from .termux import init_schema
    elif filetype in {"PKGBUILD", "install"}:
        from .pkgbuild import init_schema
    elif filetype == "makepkg.conf":
        from .makepkg_conf import init_schema
    elif filetype == "color.map":
        from .color_map import init_schema
    elif filetype == "make.conf":
        from .make_conf import init_schema
    elif filetype == "ebuild":
        from .ebuild import init_schema
    else:
        raise NotImplementedError(filetype)
    return init_schema()[filetype]



r"""PKGBUILD packages
=====================
"""
from pathlib import Path

from jinja2 import Template
from platformdirs import user_config_path
from pyalpm import Handle, Package

DB = Handle(".", "/var/lib/pacman").get_localdb()
TEMPLATE_NAME = "template.md.j2"
PATH = user_config_path("pacman") / TEMPLATE_NAME
if not PATH.exists():
    PATH = Path(__file__).parent.parent / "assets" / "jinja2" / TEMPLATE_NAME
TEMPLATE = PATH.read_text()


def render_document(pkg: Package, template: str = TEMPLATE) -> str:
    r"""Render document.

    :param pkg:
    :type pkg: Package
    :param template:
    :type template: str
    :rtype: str
    """
    return Template(template).render(pkg=pkg)


def get_package_document(name: str, template: str = TEMPLATE) -> str:
    r"""Get package document.

    :param name:
    :type name: str
    :param template:
    :type template: str
    :rtype: str
    """
    return render_document(DB.get_pkg(name), template)


def get_package_names(name: str) -> dict[str, str]:
    r"""Get package names.

    :param name:
    :type name: str
    :rtype: dict[str, str]
    """
    return {
        pkg.name: render_document(pkg)
        for pkg in DB.search(name)
        if pkg.name.startswith(name)
    }



r"""Server
==========
"""
import re
from typing import Any

from lsprotocol.types import (
    TEXT_DOCUMENT_COMPLETION,
    TEXT_DOCUMENT_DID_CHANGE,
    TEXT_DOCUMENT_DID_OPEN,
    TEXT_DOCUMENT_DOCUMENT_LINK,
    TEXT_DOCUMENT_FORMATTING,
    TEXT_DOCUMENT_HOVER,
    CompletionItem,
    CompletionItemKind,
    CompletionList,
    CompletionParams,
    DidChangeTextDocumentParams,
    DocumentFormattingParams,
    DocumentLink,
    DocumentLinkParams,
    Hover,
    MarkupContent,
    MarkupKind,
    TextDocumentPositionParams,
    TextEdit,
)
from pygls.server import LanguageServer
from tree_sitter_languages import get_parser
from tree_sitter_lsp.diagnose import get_diagnostics
from tree_sitter_lsp.finders import PositionFinder
from tree_sitter_lsp.format import get_text_edits

from .finders import (
    DIAGNOSTICS_FINDER_CLASSES,
    FORMAT_FINDER_CLASSES,
    CSVFinder,
    MinGWFinder,
    PackageFinder,
)
from .packages import (
    PACKAGE_VARIABLES,
    search_package_document,
    search_package_names,
)
from .utils import get_filetype, get_schema


class TermuxLanguageServer(LanguageServer):
    r"""Termux language server."""

    def __init__(self, *args: Any) -> None:
        r"""Init.

        :param args:
        :type args: Any
        :rtype: None
        """
        super().__init__(*args)
        self.trees = {}
        self.parser = get_parser("bash")

        @self.feature(TEXT_DOCUMENT_DID_OPEN)
        @self.feature(TEXT_DOCUMENT_DID_CHANGE)
        def did_change(params: DidChangeTextDocumentParams) -> None:
            r"""Did change.

            :param params:
            :type params: DidChangeTextDocumentParams
            :rtype: None
            """
            filetype = get_filetype(params.text_document.uri)
            if filetype == "":
                return None
            document = self.workspace.get_document(params.text_document.uri)
            self.trees[document.uri] = self.parser.parse(
                document.source.encode()
            )
            diagnostics = get_diagnostics(
                document.uri,
                self.trees[document.uri],
                DIAGNOSTICS_FINDER_CLASSES,
                filetype,
            )
            if document.path is not None and filetype == "PKGBUILD":
                from .tools.namcap import namcap

                diagnostics += namcap(document.path, document.source)
            self.publish_diagnostics(params.text_document.uri, diagnostics)

        @self.feature(TEXT_DOCUMENT_FORMATTING)
        def format(params: DocumentFormattingParams) -> list[TextEdit]:
            r"""Format.

            :param params:
            :type params: DocumentFormattingParams
            :rtype: list[TextEdit]
            """
            filetype = get_filetype(params.text_document.uri)
            if filetype == "":
                return []
            document = self.workspace.get_document(params.text_document.uri)
            return get_text_edits(
                document.uri,
                self.trees[document.uri],
                FORMAT_FINDER_CLASSES,
                filetype,
            )

        @self.feature(TEXT_DOCUMENT_DOCUMENT_LINK)
        def document_link(params: DocumentLinkParams) -> list[DocumentLink]:
            r"""Get document links.

            :param params:
            :type params: DocumentLinkParams
            :rtype: list[DocumentLink]
            """
            filetype = get_filetype(params.text_document.uri)
            if filetype == "":
                return []
            document = self.workspace.get_document(params.text_document.uri)
            if filetype in {"build.sh", "subpackage.sh"}:
                return CSVFinder(filetype).get_document_links(
                    document.uri,
                    self.trees[document.uri],
                    "https://github.com/termux/termux-packages/tree/master/packages/{{name}}/build.sh",
                )
            elif filetype in {"PKGBUILD", "install"}:
                if (
                    len(
                        MinGWFinder().find_all(
                            document.uri, self.trees[document.uri]
                        )
                    )
                    > 0
                ):
                    url = "https://packages.msys2.org/base/{{uni.get_text()}}"
                else:
                    url = "https://archlinux.org/packages/{{uni.get_text()}}"
                return PackageFinder().get_document_links(
                    document.uri,
                    self.trees[document.uri],
                    url,
                )
            raise NotImplementedError

        @self.feature(TEXT_DOCUMENT_HOVER)
        def hover(params: TextDocumentPositionParams) -> Hover | None:
            r"""Hover.

            :param params:
            :type params: TextDocumentPositionParams
            :rtype: Hover | None
            """
            filetype = get_filetype(params.text_document.uri)
            if filetype == "":
                return None
            document = self.workspace.get_document(params.text_document.uri)
            uni = PositionFinder(params.position).find(
                document.uri, self.trees[document.uri]
            )
            if uni is None:
                return None
            parent = uni.node.parent
            if parent is None:
                return None
            text = uni.get_text()
            _range = uni.get_range()
            # we only hover variable names and function names
            if not (
                uni.node.type == "variable_name"
                or uni.node.type == "word"
                and parent.type
                in {
                    "function_definition",
                    "command_name",
                }
            ):
                if (
                    parent.type == "array"
                    and parent.parent is not None
                    and parent.parent.children[0].text.decode()
                    in PACKAGE_VARIABLES.get(filetype, set())
                ):
                    result = search_package_document(text, filetype)
                    if result is None:
                        return None
                    return Hover(
                        MarkupContent(MarkupKind.Markdown, result), _range
                    )
                return None
            if description := (
                get_schema(filetype)
                .get("properties", {})
                .get(text, {})
                .get("description")
            ):
                return Hover(
                    MarkupContent(MarkupKind.Markdown, description),
                    _range,
                )
            for k, v in (
                get_schema(filetype).get("patternProperties", {}).items()
            ):
                if re.match(k, text):
                    return Hover(
                        MarkupContent(MarkupKind.Markdown, v["description"]),
                        _range,
                    )

        @self.feature(TEXT_DOCUMENT_COMPLETION)
        def completions(params: CompletionParams) -> CompletionList:
            r"""Completions.

            :param params:
            :type params: CompletionParams
            :rtype: CompletionList
            """
            filetype = get_filetype(params.text_document.uri)
            if filetype == "":
                return CompletionList(False, [])
            document = self.workspace.get_document(params.text_document.uri)
            uni = PositionFinder(params.position, right_equal=True).find(
                document.uri, self.trees[document.uri]
            )
            if uni is None:
                return CompletionList(False, [])
            parent = uni.node.parent
            if parent is None:
                return CompletionList(False, [])
            text = uni.get_text()
            if (
                parent.type == "array"
                and parent.parent is not None
                and parent.parent.children[0].text.decode()
                in PACKAGE_VARIABLES.get(filetype, set())
            ):
                return CompletionList(
                    False,
                    [
                        CompletionItem(
                            k,
                            kind=CompletionItemKind.Module,
                            documentation=MarkupContent(
                                MarkupKind.Markdown, v
                            ),
                            insert_text=k,
                        )
                        for k, v in search_package_names(
                            text, filetype
                        ).items()
                    ],
                )
            schema = get_schema(filetype)
            if parent.type == "array" and parent.parent is not None:
                items = (
                    schema["properties"]
                    .get(parent.parent.children[0].text.decode(), {})
                    .get("items", {})
                )
                enum = items.get(
                    "enum", items.get("oneOf", [{}])[0].get("enum", [])
                )
                return CompletionList(
                    False,
                    [
                        CompletionItem(
                            k,
                            kind=CompletionItemKind.Constant,
                            insert_text=k,
                        )
                        for k in enum
                    ],
                )
            return CompletionList(
                False,
                [
                    CompletionItem(
                        k,
                        kind=CompletionItemKind.Function
                        if v.get("const") == 0
                        else CompletionItemKind.Field
                        if v.get("type") == "array"
                        else CompletionItemKind.Variable,
                        documentation=MarkupContent(
                            MarkupKind.Markdown, v["description"]
                        ),
                        insert_text=k,
                    )
                    for k, v in (
                        schema.get("properties", {})
                        | {
                            k.lstrip("^").split("(")[0]: v
                            for k, v in schema.get(
                                "patternProperties", {}
                            ).items()
                        }
                    ).items()
                    if k.startswith(text)
                ],
            )



r"""Termux
==========
"""
from typing import Any

from tree_sitter_lsp.misc import get_soup

from .. import CSV
from .._metainfo import SOURCE, project
from .licenses import ATOM

URIS = {
    "variable": "https://github.com/termux/termux-packages/wiki/Creating-new-package",
    "function": "https://github.com/termux/termux-packages/wiki/Building-packages",
    "update": "https://github.com/termux/termux-packages/wiki/Auto-updating-packages",
}


def init_schema() -> dict[str, dict[str, Any]]:
    r"""Init schema.

    :rtype: dict[str, dict[str, Any]]
    """
    schemas = {}
    for filetype in {"build.sh", "subpackage.sh"}:
        schemas[filetype] = {
            "$id": f"{SOURCE}/blob/main/src/termux_language_server/assets/json/{filetype}.json",
            "$schema": "http://json-schema.org/draft-07/schema#",
            "$comment": (
                "Don't edit this file directly! It is generated by "
                f"`{project} --generate-schema={filetype}`."
            ),
            "type": "object",
            # required variables
            "required": [],
            # all variables and all overridable functions
            "properties": {},
            "propertyNames": {
                # not overridable functions
                # for subpackage.sh, disable all functions
                "not": {"anyOf": []}
            },
        }

    soup = get_soup(URIS["variable"])
    tables = soup.find_all("table")
    for tr in tables[0].find_all("tr")[1:]:
        tds = tr.find_all("td")
        name = tds[1].text
        is_required = tds[2].text == "yes"
        description = tds[3].text.strip()
        if is_required:
            schemas["build.sh"]["required"] += [name]
        schemas["build.sh"]["properties"][name] = {
            "type": "string",
            "description": description,
        }
        if description.endswith("Default is true."):
            schemas["build.sh"]["properties"][name]["default"] = "true"
            schemas["build.sh"]["properties"][name]["enum"] = ["true", "false"]
        if description.endswith("Default is false."):
            schemas["build.sh"]["properties"][name]["default"] = "false"
            schemas["build.sh"]["properties"][name]["enum"] = ["true", "false"]
        if description.startswith("Comma-separated list of "):
            schemas["build.sh"]["properties"][name]["pattern"] = CSV
    for tr in tables[2].find_all("tr")[1:]:
        tds = tr.find_all("td")
        name = tds[1].text
        is_required = tds[2].text == "yes"
        description = tds[3].text.strip()
        if is_required:
            schemas["subpackage.sh"]["required"] += [name]
        schemas["subpackage.sh"]["properties"][name] = {
            "type": "string",
            "description": description,
        }
        if description.endswith("Default is true."):
            schemas["subpackage.sh"]["properties"][name]["default"] = "true"
            schemas["subpackage.sh"]["properties"][name]["enum"] = [
                "true",
                "false",
            ]
        if description.endswith("Default is false."):
            schemas["subpackage.sh"]["properties"][name]["default"] = "false"
            schemas["subpackage.sh"]["properties"][name]["enum"] = [
                "true",
                "false",
            ]
        if description.startswith("Comma-separated list of "):
            schemas["subpackage.sh"]["properties"][name]["pattern"] = CSV
    for li in soup.find_all("ul")[-2].find_all("li"):
        name = li.find("code").text
        description = (
            li.text.partition("-")[2].strip().replace("\n", " ").strip()
        )
        obj = {
            "const": name,
            "description": description,
        }
        schemas["build.sh"]["propertyNames"]["not"]["anyOf"] += [obj]
        schemas["subpackage.sh"]["propertyNames"]["not"]["anyOf"] += [obj]

    variable_table, function_table = get_soup(URIS["update"]).find_all("table")
    for tr in variable_table.find_all("tr")[1:]:
        tds = tr.find_all("td")
        name = tds[1].text
        is_required = tds[2].text == "yes"
        description = tds[3].text.strip()
        if is_required:
            schemas["build.sh"]["required"] += [name]
        schemas["build.sh"]["properties"][name] = {
            "type": "string",
            "description": description,
        }
        if description.startswith("Comma-separated list of "):
            schemas["build.sh"]["properties"][name]["pattern"] = CSV

    for tr in (
        get_soup(URIS["function"]).find_all("tr")[1:]
        + function_table.find_all("tr")[1:]
    ):
        tds = tr.find_all("td")
        name = tds[1].text
        # NOTE: capital
        # use `!=  "no"` to allow some conditional yes
        is_overrideable = tds[2].text.lower() != "no"
        description = tds[3].text.strip()
        if is_overrideable:
            schemas["build.sh"]["properties"][name] = {
                "description": description,
                "const": 0,
            }
        else:
            obj = {
                "const": name,
                "description": description,
            }
            schemas["build.sh"]["propertyNames"]["not"]["anyOf"] += [obj]

    schemas["build.sh"]["properties"]["TERMUX_PKG_HOMEPAGE"][
        "format"
    ] = schemas["build.sh"]["properties"]["TERMUX_PKG_SRCURL"][
        "format"
    ] = "uri"
    schemas["build.sh"]["properties"]["TERMUX_PKG_MAINTAINER"][
        "default"
    ] = "@termux"
    schemas["build.sh"]["properties"]["TERMUX_PKG_UPDATE_METHOD"]["enum"] = [
        "github",
        "gitlab",
        "repology",
    ]
    schemas["build.sh"]["properties"]["TERMUX_GITLAB_API_HOST"][
        "default"
    ] = "gitlab.com"
    schemas["build.sh"]["properties"]["TERMUX_GITLAB_API_HOST"][
        "format"
    ] = "hostname"
    schemas["build.sh"]["properties"]["TERMUX_PKG_UPDATE_VERSION_REGEXP"][
        "format"
    ] = "regex"
    schemas["build.sh"]["properties"]["TERMUX_PKG_LICENSE"][
        "pattern"
    ] = rf"{ATOM}(,{ATOM})*"
    return schemas



r"""makepkg.conf
================
"""
from typing import Any

from tree_sitter_lsp.misc import get_soup

from .._metainfo import SOURCE, project


def init_schema() -> dict[str, dict[str, Any]]:
    r"""Init schema.

    :rtype: dict[str, dict[str, Any]]
    """
    filetype = "makepkg.conf"
    schema = {
        "$id": f"{SOURCE}/blob/main/src/termux_language_server/assets/json/{filetype}.json",
        "$schema": "http://json-schema.org/draft-07/schema#",
        "$comment": (
            "Don't edit this file directly! It is generated by "
            f"`{project} --generate-schema={filetype}`."
        ),
        "type": "object",
        "properties": {},
    }
    name = ""
    for blockquote in get_soup("makepkg.conf").find_all("blockquote"):
        if blockquote.find("p").contents == ["."]:
            break
        p = blockquote.find_previous("p")
        if p.find("strong") is None:
            continue
        if p.text.islower():
            schema["properties"][name]["items"]["enum"] = schema["properties"][
                name
            ]["items"].get("enum", []) + [p.text, f"!{p.text}"]
            continue
        _description = "\n".join(
            content.text.replace("\n", " ")
            for p in blockquote.find_all("p")
            for content in p.contents
        )
        last_name = ""
        example = ""
        text = ""
        name = ""
        for text in [content.text for content in p.contents]:
            # FIXME: typo of makepkg.conf
            if "=" in text or text == "COMPRESSLZO":
                if text == "COMPRESSLZO":
                    text += "="
                name = text.split("=")[0]
                if last_name != "":
                    description = f"""```sh
{example}
```
{_description}"""
                    schema["properties"][last_name] = {
                        "description": description,
                        "type": "array" if "(" in example else "string",
                    }
                    if schema["properties"][last_name]["type"] == "array":
                        schema["properties"][last_name] |= {
                            "items": {"type": "string"},
                            "uniqueItems": True,
                        }
                    example = ""
                last_name = name
            example += text.replace("\n", " ")
        else:
            description = f"""```sh
{example}
```
{_description}"""
            schema["properties"][name] = {
                "description": description,
                "type": "array" if "(" in example else "string",
            }
            if schema["properties"][last_name]["type"] == "array":
                schema["properties"][last_name] |= {
                    "items": {"type": "string"},
                    "uniqueItems": True,
                }
    return {filetype: schema}



r"""Provide ``__version__`` for
`importlib.metadata.version() <https://docs.python.org/3/library/importlib.metadata.html#distribution-versions>`_.
"""
from typing import Literal

try:
    from ._version import __version__, __version_tuple__  # type: ignore
except ImportError:  # for setuptools-generate
    __version__ = "rolling"
    __version_tuple__ = (0, 0, 0, __version__, "")

__all__ = ["__version__", "__version_tuple__"]

FILETYPE = Literal[
    "build.sh",
    "subpackage.sh",
    "PKGBUILD",
    "makepkg.conf",
    "install",
    "ebuild",
    "make.conf",
    "color.map",
]
PACKAGE_NAME = r"[a-z][a-z0-9-]*"
CSV = f"{PACKAGE_NAME}(, {PACKAGE_NAME})*"



r"""Namcap
==========
"""
from lsprotocol.types import Diagnostic, DiagnosticSeverity, Position, Range


def namcap(path: str, source: str) -> list[Diagnostic]:
    r"""Namcap.

    :param path:
    :type path: str
    :param source:
    :type source: str
    :rtype: list[Diagnostic]
    """
    try:
        from Namcap.rules import all_rules
    except ImportError:
        return []
    from Namcap.package import load_from_pkgbuild
    from Namcap.ruleclass import PkgbuildRule
    from Namcap.tags import format_message

    pkginfo = load_from_pkgbuild(path)
    items = {}
    for value in all_rules.values():
        rule = value()
        if isinstance(rule, PkgbuildRule):
            rule.analyze(pkginfo, "PKGBUILD")  # type: ignore
        for msg in rule.errors:
            items[format_message(msg)] = DiagnosticSeverity.Error
        for msg in rule.warnings:
            items[format_message(msg)] = DiagnosticSeverity.Warning
    end = len(source.splitlines()[0])
    return [
        Diagnostic(Range(Position(0, 0), Position(0, end)), msg, severity)
        for msg, severity in items.items()
    ]



r"""Portage's color.map
=======================
"""
from typing import Any

from tree_sitter_lsp.misc import get_soup

from .._metainfo import SOURCE, project


def init_schema() -> dict[str, dict[str, Any]]:
    r"""Init schema.

    :rtype: dict[str, dict[str, Any]]
    """
    filetype = "color.map"
    schema = {
        "$id": f"{SOURCE}/blob/main/src/termux_language_server/assets/json/{filetype}.json",
        "$schema": "http://json-schema.org/draft-07/schema#",
        "$comment": (
            "Don't edit this file directly! It is generated by "
            f"`{project} --generate-schema={filetype}`."
        ),
        "type": "object",
        "properties": {},
    }
    dl = get_soup("color.map").find_all("dl")[1]
    for dt, dd in zip(dl.find_all("dt"), dl.find_all("dd")):
        name = dt.text.split()[0]
        description = dd.text.replace("\n", " ").strip()
        example = dt.text.replace("\n", " ")
        if name != example:
            description = f"""```sh
{example}
```
{description}"""
        schema["properties"][name] = {
            "description": description,
            "type": "string",
        }
    return {filetype: schema}



r"""Portage's ebuild
====================
"""
from typing import Any

from tree_sitter_lsp.misc import get_soup

from .._metainfo import SOURCE, project
from .licenses import ATOM


def init_schema() -> dict[str, dict[str, Any]]:
    r"""Init schema.

    :rtype: dict[str, dict[str, Any]]
    """
    filetype = "ebuild"
    schema = {
        "$id": f"{SOURCE}/blob/main/src/termux_language_server/assets/json/{filetype}.json",
        "$schema": "http://json-schema.org/draft-07/schema#",
        "$comment": (
            "Don't edit this file directly! It is generated by "
            f"`{project} --generate-schema={filetype}`."
        ),
        "type": "object",
        "properties": {},
    }
    for dl in get_soup("ebuild").find_all("dl")[20:-2]:
        for dt, dd in zip(dl.find_all("dt"), dl.find_all("dd")):
            if dt.strong is None or dt.strong.text.endswith(":"):
                continue
            name = dt.strong.text.split()[0]
            description = dd.text.replace("\n", " ").strip()
            example = dt.text.replace("\n", " ")
            if name != example:
                description = f"""```sh
{example}
```
{description}"""
            schema["properties"][name] = {
                "description": description,
            }
            if name.isupper():
                schema["properties"][name]["type"] = "string"
            else:
                schema["properties"][name]["const"] = 0
    schema["properties"]["LICENSE"]["pattern"] = rf"{ATOM}(( |\n){ATOM})*"
    return {filetype: schema}



r"""Test schema."""
import os

from tree_sitter_languages import get_parser
from tree_sitter_lsp.finders import SchemaFinder

from termux_language_server.schema import BashTrie
from termux_language_server.utils import get_filetype, get_schema

PATH = os.path.dirname(__file__)


class Test:
    r"""Test."""

    @staticmethod
    def test_SchemaFinder() -> None:
        r"""Test schemafinder.

        :rtype: None
        """
        path = os.path.join(PATH, "build.sh")
        filetype = get_filetype(path)
        assert filetype == "build.sh"
        with open(path, "rb") as f:
            text = f.read()
        tree = get_parser("bash").parse(text)
        finder = SchemaFinder(get_schema(filetype), BashTrie)
        diagnostics = finder.get_diagnostics(path, tree)
        assert len(diagnostics) > 0



r"""Finders
===========
"""
from copy import deepcopy
from dataclasses import dataclass

from jinja2 import Template
from lsprotocol.types import (
    CompletionItemKind,
    DiagnosticSeverity,
    DocumentLink,
    Position,
    Range,
    TextEdit,
)
from tree_sitter import Node, Tree
from tree_sitter_lsp import UNI, Finder
from tree_sitter_lsp.finders import (
    ErrorQueryFinder,
    QueryFinder,
    SchemaFinder,
    UnFixedOrderFinder,
)

from . import CSV, FILETYPE
from .schema import BashTrie
from .utils import get_query, get_schema


@dataclass(init=False)
class ErrorBashFinder(ErrorQueryFinder):
    r"""Errorbashfinder."""

    def __init__(
        self,
        message: str = "{{uni.get_text()}}: error",
        severity: DiagnosticSeverity = DiagnosticSeverity.Error,
    ) -> None:
        r"""Init.

        :param filetype:
        :type filetype: str
        :param message:
        :type message: str
        :param severity:
        :type severity: DiagnosticSeverity
        :rtype: None
        """
        super().__init__("bash", message, severity)


@dataclass(init=False)
class BashFinder(SchemaFinder):
    r"""Bashfinder."""

    def __init__(self, filetype: FILETYPE) -> None:
        r"""Init.

        :param filetype:
        :type filetype: FILETYPE
        :rtype: None
        """
        self.validator = self.schema2validator(get_schema(filetype))
        self.cls = BashTrie


@dataclass(init=False)
class UnsortedKeywordFinder(UnFixedOrderFinder):
    r"""Unsortedkeywordfinder."""

    def __init__(
        self,
        filetype: FILETYPE,
        message: str = "{{uni.get_text()}}: is unsorted due to {{_uni}}",
        severity: DiagnosticSeverity = DiagnosticSeverity.Warning,
    ) -> None:
        r"""Init.

        :param filetype:
        :type filetype: FILETYPE
        :param message:
        :type message: str
        :param severity:
        :type severity: DiagnosticSeverity
        :rtype: None
        """
        super().__init__(
            list(get_schema(filetype).get("properties", [])), message, severity
        )
        self.keywords = UnsortedKeywordFinder.get_keywords(filetype)

    @staticmethod
    def get_keywords(filetype) -> dict[str, CompletionItemKind]:
        r"""Get keywords.

        :param filetype:
        :rtype: dict[str, CompletionItemKind]
        """
        return {
            k: (
                CompletionItemKind.Function
                if v.get("const") == 0
                else CompletionItemKind.Variable
            )
            for k, v in get_schema(filetype).get("properties", {}).items()
        }

    @staticmethod
    def is_correct_declaration(uni: UNI, _type: CompletionItemKind) -> bool:
        r"""Is correct declaration.

        :param uni:
        :type uni: UNI
        :param _type:
        :type _type: CompletionItemKind
        :rtype: bool
        """
        parent = uni.node.parent
        if parent is None:
            return False
        return (
            _type == CompletionItemKind.Variable
            and uni.node.type == "variable_name"
            and parent.type == "variable_assignment"
            and parent.children[-1].type != "array"
            or _type == CompletionItemKind.Field
            and uni.node.type == "variable_name"
            and parent.type == "variable_assignment"
            and parent.children[-1].type == "array"
            or _type == CompletionItemKind.Function
            and uni.node.type == "word"
            and parent.type == "function_definition"
        )

    @staticmethod
    def is_correct_reference(uni: UNI, _type: CompletionItemKind) -> bool:
        r"""Is correct reference.

        :param uni:
        :type uni: UNI
        :param _type:
        :type _type: CompletionItemKind
        :rtype: bool
        """
        parent = uni.node.parent
        if parent is None:
            return False
        return (
            _type in {CompletionItemKind.Variable, CompletionItemKind.Field}
            and uni.node.type == "variable_name"
            and parent.type in {"expansion", "simple_expansion"}
            or _type == CompletionItemKind.Function
            and uni.node.type == "word"
            and parent.type == "command_name"
        )

    @staticmethod
    def is_correct(uni: UNI, _type: CompletionItemKind) -> bool:
        r"""Is correct.

        :param uni:
        :type uni: UNI
        :param _type:
        :type _type: CompletionItemKind
        :rtype: bool
        """
        return UnsortedKeywordFinder.is_correct_declaration(
            uni, _type
        ) or UnsortedKeywordFinder.is_correct_reference(uni, _type)

    def filter(self, uni: UNI) -> bool:
        r"""Filter.

        :param uni:
        :type uni: UNI
        :rtype: bool
        """
        text = uni.get_text()
        return (
            text in self.order
            and text in self.keywords
            and UnsortedKeywordFinder.is_correct_declaration(
                uni, self.keywords[text]
            )
        )

    def get_text_edits(self, uri: str, tree: Tree) -> list[TextEdit]:
        r"""Get text edits. Only return two to avoid `Overlapping edit`

        :param self:
        :param uri:
        :type uri: str
        :param tree:
        :type tree: Tree
        :rtype: list[TextEdit]
        """
        self.find_all(uri, tree)
        for uni, _uni in self.uni_pairs:
            parent = uni.node.parent
            _parent = _uni.node.parent
            if parent is None or _parent is None:
                return []
            # swap 2 unis
            return [
                TextEdit(UNI.node2range(parent), UNI.node2text(_parent)),
                TextEdit(UNI.node2range(_parent), UNI.node2text(parent)),
            ]
        return []


@dataclass(init=False)
class UnsortedCSVFinder(Finder):
    r"""Unsorted comma separated value finder."""

    def __init__(
        self,
        filetype: FILETYPE,
        message: str = "{{uni.get_text()}}: unsorted",
        severity: DiagnosticSeverity = DiagnosticSeverity.Warning,
    ) -> None:
        r"""Init.

        :param filetype:
        :type filetype: FILETYPE
        :param message:
        :type message: str
        :param severity:
        :type severity: DiagnosticSeverity
        :rtype: None
        """
        super().__init__(message, severity)
        self.csvs = self.get_csvs(filetype)

    @staticmethod
    def get_csvs(filetype: FILETYPE) -> set[str]:
        r"""Get csvs.

        :param filetype:
        :type filetype: FILETYPE
        :rtype: set[str]
        """
        return set(
            k
            for k, v in get_schema(filetype).get("properties", {}).items()
            if v.get("pattern") == CSV
        )

    def __call__(self, uni: UNI) -> bool:
        r"""Call.

        :param uni:
        :type uni: UNI
        :rtype: bool
        """
        return self.is_csv(uni) and self.sort(uni.get_text()) != uni.get_text()

    def is_csv(self, uni: UNI) -> bool:
        r"""Is csv.

        :param uni:
        :type uni: UNI
        :rtype: bool
        """
        parent = uni.node.parent
        if parent is None:
            return False
        return (
            parent.type == "variable_assignment"
            and uni.node == parent.children[-1]
            and (uni.node.type == "word" or uni.node.type == "string")
            and UNI.node2text(parent.children[0]) in self.csvs
        )

    @staticmethod
    def sort(text: str) -> str:
        r"""Sort.

        :param text:
        :type text: str
        :rtype: str
        """
        return (
            '"'
            + ", ".join(
                sorted(word.strip() for word in text.strip('"').split(","))
            )
            + '"'
        )

    def get_text_edits(self, uri: str, tree: Tree) -> list[TextEdit]:
        r"""Get text edits. Only return two to avoid `Overlapping edit`

        :param self:
        :param uri:
        :type uri: str
        :param tree:
        :type tree: Tree
        :rtype: list[TextEdit]
        """
        text_edits = [
            TextEdit(uni.get_range(), self.sort(uni.get_text()))
            for uni in self.find_all(uri, tree)
        ]
        return text_edits


@dataclass()
class CSVFinder(UnsortedCSVFinder):
    r"""Comma separated value finder."""

    def __post_init__(self) -> None:
        r"""Post init.

        :rtype: None
        """
        self.csvs -= {"TERMUX_PKG_BLACKLISTED_ARCHES"}

    def __call__(self, uni: UNI) -> bool:
        r"""Call.

        :param uni:
        :type uni: UNI
        :rtype: bool
        """
        return self.is_csv(uni)

    def get_document_links(
        self, uri: str, tree: Tree, template: str
    ) -> list[DocumentLink]:
        r"""Get document links.

        :param uri:
        :type uri: str
        :param tree:
        :type tree: Tree
        :param template:
        :type template: str
        :rtype: list[DocumentLink]
        """
        links = []
        for uni in self.find_all(uri, tree):
            start = list(uni.node.start_point)
            text = uni.get_text()
            if text.startswith('"'):
                text = text.strip('"')
                start[1] += 1
            end = deepcopy(start)
            for name in text.split(","):
                if name.startswith(" "):
                    name = name.lstrip(" ")
                    start[1] += 1
                    end[1] += 1
                end[1] += len(name)
                links += [
                    DocumentLink(
                        Range(Position(*start), Position(*end)),
                        Template(template).render(name=name),
                    )
                ]
                start[1] += len(name) + 1
        return links


@dataclass(init=False)
class PackageFinder(QueryFinder):
    r"""Packagefinder."""

    def __init__(
        self,
        message: str = "{{uni.get_text()}}: no such file",
        severity: DiagnosticSeverity = DiagnosticSeverity.Error,
    ) -> None:
        r"""Init.

        :param message:
        :type message: str
        :param severity:
        :type severity: DiagnosticSeverity
        :rtype: None
        """
        query = get_query("package")
        super().__init__(query, message, severity)

    def capture2uni(self, capture: tuple[Node, str], uri: str) -> UNI | None:
        r"""Capture2uni.

        :param capture:
        :type capture: tuple[Node, str]
        :param uri:
        :type uri: str
        :rtype: UNI | None
        """
        node, label = capture
        uni = UNI(uri, node)
        return uni if label == "package" else None


@dataclass(init=False)
class MinGWFinder(QueryFinder):
    r"""Mingwfinder."""

    def __init__(
        self,
        message: str = "{{uni.get_text()}}: no such file",
        severity: DiagnosticSeverity = DiagnosticSeverity.Error,
    ) -> None:
        r"""Init.

        :param message:
        :type message: str
        :param severity:
        :type severity: DiagnosticSeverity
        :rtype: None
        """
        query = get_query("mingw")
        super().__init__(query, message, severity)


DIAGNOSTICS_FINDER_CLASSES = [
    ErrorBashFinder,
    BashFinder,
    UnsortedKeywordFinder,
    UnsortedCSVFinder,
]
FORMAT_FINDER_CLASSES = [UnsortedKeywordFinder, UnsortedCSVFinder]



r"""Schema
==========
"""
from dataclasses import dataclass
from typing import Literal

from lsprotocol.types import Position, Range
from tree_sitter import Node
from tree_sitter_lsp import UNI
from tree_sitter_lsp.schema import Trie


@dataclass
class BashTrie(Trie):
    r"""Bashtrie."""

    value: dict[str, "Trie"] | list["Trie"] | str | Literal[0] = 0

    @classmethod
    def from_string_node(cls, node: Node, parent: "Trie | None") -> "Trie":
        r"""From string node.

        `<https://github.com/tree-sitter/tree-sitter-bash/issues/101>`_

        :param cls:
        :param node:
        :type node: Node
        :param parent:
        :type parent: Trie | None
        :rtype: "Trie"
        """
        if node.type == "string" and node.children == 3:
            node = node.children[1]
        text = UNI.node2text(node)
        _range = UNI.node2range(node)
        if node.type in {"string", "raw_string"} and node.children != 3:
            text = text.strip("'\"")
            _range.start.character += 1
            _range.end.character -= 1
        return cls(_range, parent, text)

    @classmethod
    def from_node(cls, node: Node, parent: "Trie | None") -> "Trie":
        r"""From node.

        :param node:
        :type node: Node
        :param parent:
        :type parent: Trie | None
        :rtype: "Trie"
        """
        string_types = {
            "word",
            "string",
            "raw_string",
            "concatenation",
            "number",
            "simple_expansion",
        }
        if node.type in string_types:
            return cls.from_string_node(node, parent)
        if node.type == "function_definition":
            return cls(UNI.node2range(node), parent, 0)
        if node.type == "variable_assignment":
            if len(node.children) < 3:
                return cls(UNI.node2range(node), parent, "")
            node = node.children[2]
            if node.type == "array":
                trie = cls(UNI.node2range(node), parent, [])
                value: list[Trie] = trie.value  # type: ignore
                trie.value = [
                    cls.from_node(child, trie)
                    for child in node.children
                    if child.type in string_types
                ]
                return trie
            if node.type in string_types:
                return cls.from_string_node(node, parent)
        if node.type == "program":
            trie = cls(Range(Position(0, 0), Position(1, 0)), parent, {})
            value: dict[str, Trie] = trie.value  # type: ignore
            for child in node.children:
                if child.type in {
                    "variable_assignment",
                    "function_definition",
                }:
                    value[UNI.node2text(child.children[0])] = cls.from_node(
                        child, trie
                    )
            return trie
        raise NotImplementedError(node.type)
